#include <stdio.h>
#include "mpi.h"




void Init_Task_Grid_demo(
                          int           Ndim,
                          int           Glb_Grid_sz[],
                          int           task_Grid_endpts[],
                          int           stride_width[],
                          void         *background,
                          MPI_Datatype  MD_type,
                          int           N_charges,
                          int          *charges_pos,
                          void         *charges_pot,

                          int          *Task_Grid_Modify,
                          void         *Task_Grid
                        )
{

#  if defined( DEBUG ) || defined( DEBUG_stdout )
#  include "debug.c.h"
   int       MD_type_sz;
#  endif
#  include "boolean.c.h"

   /*  local variables  */
   int      *task_grid_sz;
   int      *bndry_lgths;
   int      *offsets;
   int      *Nplet_idx;
   int      *Nplet_world;
   int       hyper_vol;

   int      *Grid_iptr, *bg_iptr, *Qpot_iptr;
   float    *Grid_fptr, *bg_fptr, *Qpot_fptr;
   double   *Grid_dptr, *bg_dptr, *Qpot_dptr;

   int       charge_iptr;
   int       charge_idx;

   int       idx, idir, ii;



/*  Allocate Memory for local variables  */

   task_grid_sz = ( int* ) calloc( Ndim, sizeof( int ) );
   bndry_lgths  = ( int* ) calloc( Ndim, sizeof( int ) );
   offsets      = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_idx    = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_world  = ( int* ) calloc( Ndim, sizeof( int ) );

/*  Define the boundary lengths of the task grid  */

   hyper_vol = 1;
   for ( idir = 0; idir < Ndim; idir++ )
   {

      task_grid_sz[ idir ] = task_Grid_endpts[ idir + Ndim ]
                           - task_Grid_endpts[ idir ] + 1;
      bndry_lgths[ idir ]  = task_grid_sz[ idir ]
                           + 2 * stride_width[ idir ];
      hyper_vol           *= bndry_lgths[ idir ];

      /*  offsets[] for Fortran style's implicit index [1..N]  */
      /*
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ] - 1; 
      */

      /*  offsets[] for C style's implicit index [0..N-1]  */
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ]; 

   }        /*  endif of { for ( idir = 0; idir < Ndim; idir++ ) }  */

#  if defined( DEBUG ) || defined( DEBUG_stdout )
     ierr_debug = MPI_Comm_rank( MPI_COMM_WORLD, &my_tid_debug );
#  endif

#  if defined( DEBUG ) || defined( DEBUG_stdout )
      MPI_Type_size( MD_type, &MD_type_sz );
      fprintf( stdout, "**** MD_type = %d has %d bytes\n",
               MD_type, MD_type_sz );
#  endif

#  ifdef DEBUG_stdout
      fprintf( stdout, "Task %d's array Task_Grid[] at 1 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif

   switch ( MD_type )
   {
      case MPI_INTEGER:
         Grid_iptr = (int*) Task_Grid;
         bg_iptr   = (int*) background;
         Qpot_iptr = (int*) charges_pot;
         break;
      case MPI_REAL:
         Grid_fptr = (float*) Task_Grid;
         bg_fptr   = (float*) background;
         Qpot_fptr = (float*) charges_pot;
         break;
      case MPI_DOUBLE_PRECISION:
         Grid_dptr = (double*) Task_Grid;
         bg_dptr   = (double*) background;
         Qpot_dptr = (double*) charges_pot;
         break;
      case MPI_COMPLEX:
         Grid_fptr = (float*) Task_Grid;
         bg_fptr   = (float*) background;
         Qpot_fptr = (float*) charges_pot;
         break;
      case MPI_DOUBLE_COMPLEX:
         Grid_dptr = (double*) Task_Grid;
         bg_dptr   = (double*) background;
         Qpot_dptr = (double*) charges_pot;
         break;
      default:
         fprintf( stderr, __FILE__": Wrong type in Init_Task_Grid()\n" );
         break;
   }

/*  Initialize the Task_Grid[] and Task_Grid_Modify[]  */

   for ( idx = 0; idx < hyper_vol; idx++ )
   {
      adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

      /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
      for ( ii = 0; ii < Ndim; ii++ )
         Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

      if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
      {
         switch ( MD_type )
         {
            case MPI_INTEGER:
               Grid_iptr[ idx ] = bg_iptr[ 0 ];
               break;
            case MPI_REAL:
               Grid_fptr[ idx ] = bg_fptr[ 0 ];
               break;
            case MPI_DOUBLE_PRECISION:
               Grid_dptr[ idx ] = bg_dptr[ 0 ];
               break;
            case MPI_COMPLEX:
               Grid_fptr[ 2*idx   ] = bg_fptr[ 0 ];
               Grid_fptr[ 2*idx+1 ] = bg_fptr[ 1 ];
               break;
            case MPI_DOUBLE_COMPLEX:
               Grid_dptr[ 2*idx   ] = bg_dptr[ 0 ];
               Grid_dptr[ 2*idx+1 ] = bg_dptr[ 1 ];
               break;
            default:
               fprintf( stderr, __FILE__
                        ": Unable to handle datatype %d in Init_Task_Grid()\n",
                        MD_type );
               return;
         }
         Task_Grid_Modify[ idx ] = iTRUE;
      }
      else
      {
         Task_Grid_Modify[ idx ] = iFALSE;
      }  /*  if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )  */
    
   }     /*  for ( idx = 0; idx < hyper_vol; idx++ )   */

/*   Set the Task_Grid[] and Task_Grid_Modify[] with Charges Values  */

   for ( idx = 0; idx < N_charges; idx++ )
   {
      charge_iptr = idx * Ndim;
      if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts,
                             &charges_pos[ charge_iptr ] ) )
      {
         /*  charges_pos[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
         for ( ii = 0; ii < Ndim; ii++ )
            Nplet_idx[ ii ] = charges_pos[ charge_iptr + ii ] - offsets[ ii ];

         adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_idx, &charge_idx );
         
         Task_Grid_Modify[ charge_idx ] = iFALSE;
         switch ( MD_type )
         {
            case MPI_INTEGER:
               Grid_iptr[ charge_idx ] = Qpot_iptr[ idx ];
               break;
            case MPI_REAL:
               Grid_fptr[ charge_idx ] = Qpot_fptr[ idx ];
               break;
            case MPI_DOUBLE_PRECISION:
               Grid_dptr[ charge_idx ] = Qpot_dptr[ idx ];
               break;
            case MPI_COMPLEX:
               Grid_fptr[ 2*charge_idx   ] = Qpot_fptr[ 2*idx ];
               Grid_fptr[ 2*charge_idx+1 ] = Qpot_fptr[ 2*idx+1 ];
               break;
            case MPI_DOUBLE_COMPLEX:
               Grid_dptr[ 2*charge_idx   ] = Qpot_dptr[ 2*idx ];
               Grid_dptr[ 2*charge_idx+1 ] = Qpot_dptr[ 2*idx+1 ];
               break;
            default:
               fprintf( stderr, __FILE__
                        ": Unable to handle datatype %d in Init_Task_Grid()\n",
                        MD_type );
               return;
         }
      }   /*  if (Nplet_in_RealWrd( Ndim, task_Grid_endpts, &charge_pos[] ))  */
   }

#  ifdef DEBUG
      sprintf( filename_debug, filename_debug_fmt, my_tid_debug );
      file_debug = fopen( filename_debug, "a+" );
      fprintf( file_debug, "\n" );
      print_Task_Grid( file_debug, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( file_debug, "\n" );
      fclose( file_debug );

      fprintf( stdout, "Task %d's array Task_Grid[] at 2 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif 

   free( Nplet_world );
   free( Nplet_idx );
   free( offsets );
   free( bndry_lgths );
   free( task_grid_sz );

}



#ifdef FortranSingleUnderscore
#   define init_task_grid_demo__ init_task_grid_demo_
#endif

/*  Fortran interface for RGX_Init_Task_Grid()  */
void init_task_grid_demo__(
                            int          *Ndim,
                            int          *Glb_Grid_sz,
                            int          *task_Grid_endpts,
                            int          *stride_width,
                            void         *background,
                            MPI_Datatype *MD_type,
                            int          *N_charges,
                            int          *charges_pos,
                            void         *charges_pot,

                            int          *Task_Grid_Modify,
                            void         *Task_Grid
                          )
{
   Init_Task_Grid_demo( *Ndim, Glb_Grid_sz, task_Grid_endpts, stride_width,
                        background, *MD_type, *N_charges, 
                        charges_pos, charges_pot,
                        Task_Grid_Modify, Task_Grid );
}
