#include <stdio.h>
#include "mpi.h"

void wrt_column_Task_Grid( 
                           FILE           *fout, 
                           int             Ndim, 
                           int            *task_Grid_endpts,
                           int            *stride_width, 
                           MPI_Datatype    MD_type, 
                           void           *Grid_data
                         )
{
   /*  Local variables  */
   int      *task_grid_sz;
   int      *bndry_lgths;
   int      *offsets;
   int      *Nplet_idx;
   int      *Nplet_world;
   int       hyper_vol;
   
   int      *Grid_iptr;
   float    *Grid_fptr;
   double   *Grid_dptr;

   int       elem_iptr, row_lgth;
   int       idx, idir, ii;


/*  Allocate Memory for local variables  */

   task_grid_sz = ( int* ) calloc( Ndim, sizeof( int ) );
   bndry_lgths  = ( int* ) calloc( Ndim, sizeof( int ) );
   offsets      = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_idx    = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_world  = ( int* ) calloc( Ndim, sizeof( int ) );

/*  Define the boundary lengths of the task grid  */

   hyper_vol = 1;
   for ( idir = 0; idir < Ndim; idir++ )
   {

      task_grid_sz[ idir ] = task_Grid_endpts[ idir + Ndim ]
                           - task_Grid_endpts[ idir ] + 1;
      bndry_lgths[ idir ]  = task_grid_sz[ idir ]
                           + 2 * stride_width[ idir ];
      hyper_vol           *= bndry_lgths[ idir ];

      /*  offsets[] for Fortran style's implicit index [1..N]  */
      /*
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ] - 1; 
      */

      /*  offsets[] for C style's implicit index [0..N-1]  */
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ]; 

   }        /*  endif of { for ( idir = 0; idir < Ndim; idir++ ) }  */

   row_lgth = bndry_lgths[ 0 ];

   switch ( MD_type )
   {
      case MPI_INTEGER:
         Grid_iptr = (int*) Grid_data;
         elem_iptr = 0;
         while ( elem_iptr < hyper_vol )
         {
            for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
               for ( ii = 0; ii < Ndim; ii++ )
                  Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

               if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     fprintf( fout, "% 4d ", Nplet_world[ ii ] );
                  fprintf( fout, "%d\n", Grid_iptr[ idx ] );
               }
            }  /*  for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ ) */
            fprintf( fout, "\n" );

            elem_iptr += row_lgth;
         }     /*  while ( elem_iptr < hyper_vol )  */
         break;
      case MPI_REAL:
         Grid_fptr = (float*) Grid_data;
         elem_iptr = 0;
         while ( elem_iptr < hyper_vol )
         {
            for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
               for ( ii = 0; ii < Ndim; ii++ )
                  Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

               if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     fprintf( fout, "% 4d ", Nplet_world[ ii ] );
                  fprintf( fout, "%f\n", Grid_fptr[ idx ] );
               }
            }  /*  for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ ) */
            fprintf( fout, "\n" );

            elem_iptr += row_lgth;
         }     /*  while ( elem_iptr < hyper_vol )  */
         break;
      case MPI_DOUBLE_PRECISION:
         Grid_dptr = (double*) Grid_data;
         elem_iptr = 0;
         while ( elem_iptr < hyper_vol )
         {
            for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );
 
               /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
               for ( ii = 0; ii < Ndim; ii++ )
                  Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];
 
               if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     fprintf( fout, "% 4d ", Nplet_world[ ii ] ); 
                  fprintf( fout, "%f\n", Grid_dptr[ idx ] );
               }
            }  /*  for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ ) */
            fprintf( fout, "\n" );

            elem_iptr += row_lgth;
         }     /*  while ( elem_iptr < hyper_vol )  */
         break;
      case MPI_COMPLEX:
         Grid_fptr = (float*) Grid_data;
         elem_iptr = 0;
         while ( elem_iptr < hyper_vol )
         {
            for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
               for ( ii = 0; ii < Ndim; ii++ )
                  Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

               if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     fprintf( fout, "% 4d ", Nplet_world[ ii ] );
                  fprintf( fout, "%f %f\n",
                           Grid_fptr[ 2*idx ], Grid_fptr[ 2*idx+1 ] );
               }
            }  /*  for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ ) */
            fprintf( fout, "\n" );

            elem_iptr += row_lgth;
         }     /*  while ( elem_iptr < hyper_vol )  */
         break;
      case MPI_DOUBLE_COMPLEX:
         Grid_dptr = (double*) Grid_data;
         elem_iptr = 0;
         while ( elem_iptr < hyper_vol )
         {
            for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
               for ( ii = 0; ii < Ndim; ii++ )
                  Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

               if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     fprintf( fout, "% 4d ", Nplet_world[ ii ] );
                  fprintf( fout, "%f %f\n",
                           Grid_dptr[ 2*idx ], Grid_dptr[ 2*idx+1 ] );
               }
            }  /*  for ( idx = elem_iptr; idx < elem_iptr + row_lgth; idx++ ) */
            fprintf( fout, "\n" );

            elem_iptr += row_lgth;
         }     /*  while ( elem_iptr < hyper_vol )  */
         break;
      default:
         fprintf( stderr, __FILE__": Wrong type in wrt_column_Task_Grid()\n" );
         break;
   }

   free( Nplet_world );
   free( Nplet_idx );
   free( offsets );
   free( bndry_lgths );
   free( task_grid_sz );

}  /*    ENDOF the subroutine  */
