c
c  This is a test program for 3D grid with thicker strides
c

      Program Test_3D_grid

      implicit none

      include "mpif.h"
# if defined( DEBUG ) || defined( DEBUG_stdout )
      include "debug.f.h"
# endif

      integer Ndim                        ! Dimensionally of the grid
# ifdef DIM3
      parameter ( Ndim = 3 )
# else
      parameter ( Ndim = 2 )
# endif

      integer max_sz                      ! max. size in each dimension
      parameter ( max_sz = 16 )
      integer Nelem                       ! total No. of elem in the block array
      parameter ( Nelem = max_sz * max_sz * max_sz )

      integer MD_type
      parameter ( MD_type = MPI_REAL    )

      real*4      Task_xGrid( Nelem )     ! The block array
      data Task_xGrid / Nelem * 0 /

      integer Comm                        ! Communicator of all sub-grid
      integer my_task_id                  ! tid
      integer N_task                      ! No. of process/tasks in Comm
      integer Glb_Grid_sz( Ndim )         ! Global grid sizes
      integer task_nbrs( 2, Ndim )        ! the tid of the nearest neighbor task
      integer task_Grid_endpts( Ndim, 2 ) ! local sub-grid's end points 
      integer N_task_1dims( Ndim )        ! No of task per dimension
      integer stride_width( Ndim )        ! width of the stride of ghost point
      integer task_bdy_strides( Ndim )    ! local task/grid's boundary strides
      logical periods( Ndim )             ! boolean for the periodicity of grid
      
c Dummy Variables
      character*32 node_name
      integer name_lgth
      integer swp_idx
      integer rc, ierr
      integer ii, ss



      call MPI_init( ierr )

c  Create Cartesian Regular Grids.... 

      call RGX_Create_Task_GridS( MPI_COMM_WORLD, Ndim,
     &                            N_Task, N_task_1dims,
     &                            Comm, ierr )

c Input from user to determine the size of the global grid, Glb_Grid_sz[],
c and widths of the strides, stride_width[].

      call MPI_Comm_rank( Comm, my_task_id, ierr )
      if ( my_task_id .eq. 0 ) then
         do ii = 1, Ndim
            write(6,*) 'Enter Glb_Grid_sz(',ii,') ='
            read(5,*) Glb_Grid_sz(ii)
            write(6,*) 'Enter stride_width(',ii,') ='
            read(5,*) stride_width(ii)
         enddo
      endif

c  Set/Bcast Global Grid Properties: Glb_Grid_sz[] & stride_width[]

      call RGX_Set_Glb_Grid_Ppty( Comm, Ndim, Glb_Grid_sz, stride_width,
     &                            ierr )

c Decompose the global grid into local/task sub-grid and
c Determine the end-point of the local/task sub-grid

      call RGX_Find_Task_Grid_endpts( Comm, Ndim, Glb_Grid_sz,
     &                                task_Grid_endpts, ierr )
      write(6,*) "Task ", my_task_id, "'s grid endpts: ",
     &           ( ( task_Grid_endpts( ii, ss ), ss=1,2 ), ii=1,Ndim )

c Define the "stride" for the ghost-point region

      call RGX_Defn_Task_Bndry_Strides( MD_type,
     &                                  Ndim, task_Grid_endpts,
     &                                  stride_width, task_bdy_strides,
     &                                  ierr )

c Locate my nearest neighbor sub-grid's rank for exchange of Bndry_Strides

      call RGX_Find_Task_Grid_Ngbrs( Comm, Ndim, task_nbrs, ierr )
      write(6,*) "Task ", my_task_id, "'s neighbor tid: ",
     &           ( ( task_nbrs( ss, ii ), ss=1,2 ), ii=1,Ndim )


c Clean up debugging file for C code

#     if defined( DEBUG ) || defined( DEBUG_stdout )

         if ( my_task_id .lt. 10 ) then
            write( filename_debug, fmt="( a6,i1,a4 )" ) 
     &             'node_0',my_task_id,'.txt'
         else
            write( filename_debug, fmt="( a5,i2,a4 )" ) 
     &            'node_',my_task_id,'.txt'
         endif
         call system( "rm -f "//filename_debug )

#     endif

#     ifdef DEBUG
         open( ch_debug, file=filename_debug, status='unknown' )
         rewind( ch_debug )
#     endif

c Initialize the local N-dim. grid
      call MPI_BARRIER( MPI_COMM_WORLD, ierr )

      call Init_Task_Grid_order( Ndim, Glb_Grid_sz, task_Grid_endpts,
     &                           stride_width, 1, MD_type, Task_xGrid )

      do swp_idx = 1, 2
         call RGX_Swap_Task_Bndry_Strides( Comm, Ndim,
     &                                     task_Grid_endpts,
     &                                     stride_width,
     &                                     task_bdy_strides,
     &                                     task_nbrs, MD_type,
     &                                     Task_xGrid, ierr )
         call Init_Task_Grid_order( Ndim, Glb_Grid_sz, task_Grid_endpts,
     &                              stride_width, swp_idx+1, MD_type,
     &                              Task_xGrid )

      enddo

#     ifdef DEBUG_stdout

         do ii = 1, Ndim
            task_bndry_lgths( ii ) = task_Grid_endpts( ii, 2 )
     &                             - task_Grid_endpts( ii, 1 ) + 1
     &                             + 2 * stride_width( ii )
         enddo

         write(6,*) "Fortran:  Task_Grid[] = "
         if ( MD_type .eq. MPI_INTEGER ) 
     &      call  print_Task_iGrid( 6, Ndim, task_bndry_lgths,
     &                              Task_xGrid )
         if ( MD_type .eq. MPI_REAL ) 
     &      call print_Task_r4Grid( 6, Ndim, task_bndry_lgths,
     &                              Task_xGrid )
         if ( MD_type .eq. MPI_DOUBLE_PRECISION )
     &      call print_Task_r8Grid( 6, Ndim, task_bndry_lgths,
     &                              Task_xGrid )
         if ( MD_type .eq. MPI_COMPLEX ) 
     &      call print_Task_c4Grid( 6, Ndim, task_bndry_lgths,
     &                              Task_xGrid )
         if ( MD_type .eq. MPI_DOUBLE_COMPLEX ) 
     &      call print_Task_c8Grid( 6, Ndim, task_bndry_lgths,
     &                              Task_xGrid )

#     endif

      do ii = 1, Ndim
         call MPI_Type_free( task_bdy_strides( ii ), ierr )
      enddo
      call MPI_FINALIZE(rc)

#     ifdef DEBUG
         close( ch_debug )
#     endif

      end
